# Copyright (c) 2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

from io import TextIOWrapper
from typing import Callable, List, Optional


def to_camel_case(snake_str: str) -> str:
    """Return a camel case version of a snake case string.

    Args:
        snake_str (str): A snake case string.

    Returns:
        str: A camel case version of a snake case string.
    """
    return "".join(x.capitalize() for x in snake_str.lower().split("-"))


def create_truncated_string(input: str, length: int) -> str:
    """Create a truncated version of input if it is longer than length.
    Will keep the rightmost characters and cut of the front if it is
    longer than allowed.

    Args:
        input (str): The input string.
        length (int): The allowed overall length.

    Returns:
        str: A truncated string which has len() of length.
    """
    if len(input) < length:
        return input

    return f"...{input[-length+3:]}"  # noqa: E226 intended behaviour


def replace_text_in_file(file_path: str, text: str, replacement: str) -> None:
    """Replace all occurrences of text in a file with a replacement.

    Args:
        file_path (str): The path to the file.
        text (str): The text to find.
        replacement (str): The replacement for text.
    """

    with open(file_path, mode="r+", encoding="utf-8") as file:
        file_text = file.readlines()
        replaced_text_list = replace_item_in_list(file_text, text, replacement)
        replaced_text = "".join(replaced_text_list)
        file.write(replaced_text)


def replace_item_in_list(
    text: List[str],
    matching_text: str,
    replacement: str = "",
    remove_empty: bool = False,
) -> List[str]:
    """Replace the whole line which matches the given text with a replacement.

    Args:
        text (List[str]): All text lines to replace lines within.
        text (str): The text to find the line with.
        replacement (str): The replacement for line.
        remove_empty (bool): If true and the replacement for a line is empty, the line will be removed.
    """
    buffer: List[str] = []
    for line in text:
        if matching_text in line:
            line = replacement

            is_removing_empty_line = remove_empty and line == ""
            if is_removing_empty_line:
                continue

        buffer.append(line)

    return buffer


def replace_text_area(
    text: List[str], start_occurence: str, end_occurence: str, replacement: str = ""
) -> List[str]:
    """Replace all occurrences of all text areas matching the parameters with a replacement.
    If the replacement for a line is empty, then the line will be removed.

    Args:
        text (List[str]): All text lines to replace text within.
        start_occurence (str): The starting line which matches the occurence for replacement text area.
        start_occurence (str): The ending line which matches the occurence for replacement text area.
        replacement (str): The replacement for text area.
    """
    buffer = []
    is_capturing = False

    for line in text:
        if not is_capturing:
            if start_occurence in line:
                is_capturing = True
                continue

            buffer.append(line)
            continue

        if end_occurence in line:
            is_capturing = False

            if replacement:
                buffer.append(replacement)

    return buffer


def capture_area_in_file(
    file: TextIOWrapper,
    start_line: str,
    end_line: str,
    map_fn: Optional[Callable[[str], str]] = None,
) -> List[str]:
    """Capture an area of a textfile between a matching start line (exclusive) and the first line matching end_line (exclusive).

    Args:
        file (TextIOWrapper): The text file to read from.
        start_line (str): The line which triggers the capture (will not be part of the output)
        end_line (str): The line which terminates the capture (will not be bart of the output)
        map_fn (Optional[Callable[[str], str]], optional): An optional mapping function to transform captured lines. Defaults to None.

    Returns:
        List[str]: A list of captured lines.
    """
    area_content: List[str] = []
    is_capturing = False
    for line in file:
        if line.strip() == start_line:
            is_capturing = True
        elif line.strip() == end_line:
            is_capturing = False
        elif is_capturing:
            line = line.rstrip()

            if map_fn:
                line = map_fn(line)

            area_content.append(line)
    return area_content



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class MiddlewareType(Enum):
    """Enumeration containing all possible middleware types."""

    NATIVE = 0


def get_middleware_type() -> MiddlewareType:
    """Return the current middleware type."""
    return MiddlewareType.NATIVE



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

# flake8: noqa: U100 unused argument (because of pytest.fixture)

import json
import os

import pytest

from velocitas_lib import get_cache_data, get_package_path
from velocitas_lib.variables import ProjectVariables, json_obj_to_flat_map


@pytest.fixture()
def set_test_env_var():
    os.environ["VELOCITAS_CACHE_DATA"] = '{"cache_key":"cache_value"}'


@pytest.fixture()
def set_velocitas_cache_data() -> str:
    cache_data_mock = {"testPropA": "testValueA", "testPropB": "testValueB"}
    os.environ["VELOCITAS_CACHE_DATA"] = json.dumps(cache_data_mock)
    return os.environ["VELOCITAS_CACHE_DATA"]


def test_replace_occurrences__returns_correct_resolved_string(set_test_env_var):
    input_str_a = "${{ test.string.a }}"
    input_str_b = "/test/${{ test.string.b }}/test"

    variables_to_replace = {
        "test.string.a": "testA",
        "test.string.b": "testB",
    }

    variables = ProjectVariables(variables_to_replace)
    assert (
        variables.replace_occurrences(input_str_a)
        == variables_to_replace["test.string.a"]
    )
    assert (
        variables.replace_occurrences(input_str_b)
        == f'/test/{variables_to_replace["test.string.b"]}/test'
    )


def test_replace_occurrences__variable_not_defined__raises_KeyError(set_test_env_var):
    with pytest.raises(KeyError):
        input_str_a = "${{ test.string.a }}"
        variables_to_replace = {
            "test.string.b": "testB",
        }
        ProjectVariables(variables_to_replace).replace_occurrences(input_str_a)


def test_replace_occurrences__no_replacement_in_input_str__returns_input_str(
    set_test_env_var,
):
    input_str_a = "test.string.a"
    input_str_b = "/test/test.string.b/test"
    input_str_c = "testImage:testVersion"
    input_str_d = "url.com/owner/repo/service:version"
    variables_to_replace = {
        "test.string.a": "testA",
        "test.string.b": "testB",
    }

    variables = ProjectVariables(variables_to_replace)
    assert variables.replace_occurrences(input_str_a) == input_str_a
    assert variables.replace_occurrences(input_str_b) == input_str_b
    assert variables.replace_occurrences(input_str_c) == input_str_c
    assert variables.replace_occurrences(input_str_d) == input_str_d


def test_replace_occurrences__builtins_provided(set_test_env_var):
    variables = ProjectVariables({})

    assert (
        variables.replace_occurrences(
            "This is the package path: '${{ builtin.package_dir }}'"
        )
        == f"This is the package path: {get_package_path().__str__()!r}"
    )


def test_json_obj_to_flat_map__obj_is_dict__returns_replaced_cache_data_with_separator(
    set_velocitas_cache_data,  # type: ignore
):
    separator = "test.separator"
    cache_data_with_keys_to_replace = json_obj_to_flat_map(get_cache_data(), separator)
    assert cache_data_with_keys_to_replace[f"{separator}.testPropA"] == "testValueA"
    assert cache_data_with_keys_to_replace[f"{separator}.testPropB"] == "testValueB"


def test_json_obj_to_flat_map__obj_is_list__returns_replaced_cache_data_with_separator(
    set_velocitas_cache_data,  # type: ignore
):
    separator = "test.separator"
    cache_data_with_keys_to_replace = json_obj_to_flat_map(
        list(get_cache_data()), separator
    )
    assert cache_data_with_keys_to_replace[f"{separator}.0"] == "testPropA"
    assert cache_data_with_keys_to_replace[f"{separator}.1"] == "testPropB"


def test_json_obj_to_flat_map__obj_is_str__returns_replaced_cache_data_with_separator(
    set_velocitas_cache_data,  # type: ignore
):
    separator = "test.separator"
    cache_data_with_keys_to_replace = json_obj_to_flat_map("test", separator)
    assert cache_data_with_keys_to_replace[f"{separator}"] == "test"



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

import json
import os
import sys
import re
import requests
from io import TextIOWrapper
from typing import Any, Dict, Optional


def get_valid_arch(arch: str) -> str:
    """Return a known architecture for the given `arch`.

    Args:
        arch (str): The architecture of the profile.

    Returns:
        str: valid architecture.
    """
    if "x86_64" in arch or "amd64" in arch:
        return "x86_64"
    elif "aarch64" in arch or "arm64" in arch:
        return "aarch64"

    raise ValueError(f"Unknown architecture: {arch}")


def require_env(name: str) -> str:
    """Require and return an environment variable.

    Args:
        name (str): The name of the variable.

    Raises:
        ValueError: In case the environment variable is not set.

    Returns:
        str: The value of the variable.
    """
    var = os.getenv(name)
    if not var:
        raise ValueError(f"Environment variable {name!r} not set!")
    return var


def get_workspace_dir() -> str:
    """Return the workspace directory."""
    return require_env("VELOCITAS_WORKSPACE_DIR")


def get_app_manifest() -> Dict[str, Any]:
    manifest_data = json.loads(require_env("VELOCITAS_APP_MANIFEST"))
    if isinstance(manifest_data, dict):
        return manifest_data
    elif isinstance(manifest_data, list) and isinstance(manifest_data[0], dict):
        return manifest_data[0]
    else:
        raise TypeError("Manifest must be a dict or array!")


def get_script_path() -> str:
    """Return the absolute path to the directory the invoked Python script
    is located in."""
    return os.path.dirname(os.path.realpath(sys.argv[0]))


def get_package_path() -> str:
    """Return the absolute path to the package directory the invoked Python
    script belongs to."""
    return require_env("VELOCITAS_PACKAGE_DIR")


def get_project_cache_dir() -> str:
    """Return the project's cache directory.

    Returns:
        str: The path to the project's cache directory.
    """
    return require_env("VELOCITAS_CACHE_DIR")


def get_cache_data() -> Dict[str, Any]:
    """Return the data of the cache as Python object."""
    cache_data = json.loads(require_env("VELOCITAS_CACHE_DATA"))

    if not isinstance(cache_data, dict):
        raise TypeError("VELOCITAS_CACHE_DATA has to be a JSON object!")

    return cache_data


def get_log_file_name(service_id: str, runtime_id: str) -> str:
    """Build the log file name for the given service and runtime.

    Args:
        service_id (str): The ID of the service to log.
        runtime_id (str): The ID of the runtime to log.

    Returns:
        str: The log file name.
    """
    return os.path.join(get_workspace_dir(), "logs", runtime_id, f"{service_id}.log")


def get_programming_language() -> str:
    """Return the programming language of the project."""
    return require_env("language")


def create_log_file(service_id: str, runtime_id: str) -> TextIOWrapper:
    """Create a log file for the given service and runtime.

    Args:
        service_id (str): The ID of the service to log.
        runtime_id (str): The ID of the runtime to log.

    Returns:
        TextIOWrapper: The log file.
    """
    log_file_name = get_log_file_name(service_id, runtime_id)
    os.makedirs(os.path.dirname(log_file_name), exist_ok=True)
    return open(log_file_name, "w", encoding="utf-8")


def download_file(uri: str, local_file_path: str) -> None:
    with requests.get(uri, timeout=30) as infile:
        os.makedirs(os.path.split(local_file_path)[0], exist_ok=True)
        with open(local_file_path, "wb") as outfile:
            for chunk in infile.iter_content(chunk_size=8192):
                outfile.write(chunk)


def is_uri(path: str) -> bool:
    """Check if the provided path is a URI.

    Args:
        path (str): The path to check.

    Returns:
        bool: True if the path is a URI. False otherwise.
    """
    return re.match(r"(\w+)\:\/\/(\w+)", path) is not None


def obtain_local_file_path(
    path_or_uri: str, download_path: Optional[str] = None
) -> str:
    """Return the absolute path to the file, specified by a absolute/relative local path or with an URI.

    Args:
        path_or_uri (str): Absolute/relative local path or URI.
        download_path (str): The path to download the file.

    Returns:
        str: The absolute path to the file.
    """
    if not is_uri(path_or_uri):
        if os.path.isfile(path_or_uri):
            return path_or_uri
        elif os.path.isfile(os.path.join(get_workspace_dir(), path_or_uri)):
            return os.path.join(get_workspace_dir(), path_or_uri)
        else:
            raise FileNotFoundError(f"File {path_or_uri} not found!")

    if download_path is None:
        download_path = os.path.join(
            get_project_cache_dir(), "downloads", path_or_uri.split("/")[-1]
        )
    if os.path.isfile(download_path):
        path, file = os.path.split(download_path)
        parts = file.split(".", 1)
        filename = f"{parts[0]}_1.{parts[1]}" if len(parts) > 1 else f"{parts[0]}_1"

        download_path = os.path.join(path, filename)

    download_file(path_or_uri, download_path)
    return download_path



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

import os
import sys

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem
from velocitas_lib import get_workspace_dir

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "src"))
from velocitas_lib.conan_utils import add_dependency_to_conanfile  # noqa


@pytest.fixture
def env():
    os.environ["VELOCITAS_WORKSPACE_DIR"] = "/home/workspace"


def test_add_dependency_to_conanfile__only_requires_section(fs: FakeFilesystem, env):
    contents = """
[requires]
"""

    conanfile_path = os.path.join(get_workspace_dir(), "conanfile.txt")
    fs.create_file(conanfile_path, contents=contents)
    add_dependency_to_conanfile("mydep", "myver")

    expected = """
[requires]
mydep/myver
"""

    assert expected == open(conanfile_path, encoding="utf-8").read()


def test_add_dependency_to_conanfile__multiple_sections(fs: FakeFilesystem, env):
    contents = """
[requires]

[foo]

[bar]
"""

    conanfile_path = os.path.join(get_workspace_dir(), "conanfile.txt")
    fs.create_file(conanfile_path, contents=contents)
    add_dependency_to_conanfile("mydep", "myver")

    expected = """
[requires]
mydep/myver

[foo]

[bar]
"""

    assert expected == open(conanfile_path, encoding="utf-8").read()


def test_add_dependency_to_conanfile__no_requires_section(fs: FakeFilesystem, env):
    contents = """
[foo]

[bar]
"""

    conanfile_path = os.path.join(get_workspace_dir(), "conanfile.txt")
    fs.create_file(conanfile_path, contents=contents)
    add_dependency_to_conanfile("mydep", "myver")

    expected = """
[foo]

[bar]
[requires]
mydep/myver
"""

    assert expected == open(conanfile_path, encoding="utf-8").read()


def test_add_dependency_to_conanfile__pre_existing_dep_(fs: FakeFilesystem, env):
    contents = """
[foo]

[requires]
mydep/myver2

[bar]
"""

    conanfile_path = os.path.join(get_workspace_dir(), "conanfile.txt")
    fs.create_file(conanfile_path, contents=contents)
    add_dependency_to_conanfile("mydep", "myver")

    expected = """
[foo]

[requires]
mydep/myver

[bar]
"""

    assert expected == open(conanfile_path, encoding="utf-8").read()



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

import os
import subprocess
from typing import List, Optional, Tuple

from velocitas_lib import get_workspace_dir


def get_required_sdk_version() -> Optional[str]:
    """Return the required version of the core SDK.

    Returns:
        Optional[str]: The required version or None in case SDK is not a dependency.
    """
    sdk_version: Optional[str] = None
    with open(
        os.path.join(get_workspace_dir(), "conanfile.txt"), encoding="utf-8"
    ) as conanfile:
        for line in conanfile:
            if line.startswith("vehicle-app-sdk/"):
                sdk_version = line.split("/", maxsplit=1)[1].split("@")[0].strip()

    return sdk_version


def export_conan_project(conan_project_path: str) -> None:
    """Export a conan project to the local conan cache.

    Args:
        conan_project_path (str): The path to directory containing the project.
    """
    env = os.environ.copy()
    env["CONAN_REVISIONS_ENABLED"] = "1"
    print("Exporting Conan project")
    subprocess.check_call(
        ["conan", "export", "."],
        cwd=conan_project_path,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        env=env,
    )


def _find_insertion_index(
    lines: List[str], dependency_name: str
) -> Tuple[int, bool, bool]:
    """Find an insertion index for the dependency in a conanfile.txt.

    Args:
        lines (List[str]): The lines of the original conanfile.txt
        dependency_name (str): The name of the dependency (without version) e.g. "grpc"
            of the dependency to insert.

    Returns:
        Tuple[int, bool, bool]: A tuple consisting of
            [0] = Insert index.
            [1] = Whether the insert index replaces the original line or not.
            [2] = Whether the original file has a requires section or not.
    """
    found_index: Optional[int] = None
    replace: bool = False
    in_requires_section = False
    has_requires_section = False
    for i in range(0, len(lines)):
        stripped_line = lines[i].strip()
        if stripped_line == "[requires]":
            has_requires_section = True
            in_requires_section = True
            found_index = i + 1
        elif in_requires_section and stripped_line.startswith("["):
            in_requires_section = False

        if in_requires_section:
            if len(stripped_line) > 0:
                if stripped_line.startswith(dependency_name):
                    found_index = i
                    replace = True

    if found_index is None:
        found_index = len(lines)

    return (found_index, replace, has_requires_section)


def add_dependency_to_conanfile(dependency_name: str, dependency_version: str) -> None:
    """Add the dependency name to the project's list of dependencies.

    Args:
        dependency_name (str): The dependency to add e.g. grpc
        dependency_version (str): The version of the dependency to add e.g. 1.50.1
    """
    conanfile_path = os.path.join(get_workspace_dir(), "conanfile.txt")

    lines = []
    with open(conanfile_path, encoding="utf-8", mode="r") as conanfile:
        lines = conanfile.readlines()

    insert_index, replace, has_requires_section = _find_insertion_index(
        lines, dependency_name
    )

    dependency_line = f"{dependency_name}/{dependency_version}\n"
    if replace:
        lines[insert_index] = dependency_line
    else:
        if not has_requires_section:
            lines.insert(insert_index, "[requires]\n")
            insert_index = insert_index + 1
        lines.insert(insert_index, dependency_line)

    with open(conanfile_path, encoding="utf-8", mode="w") as conanfile:
        conanfile.writelines(lines)



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

import os
from typing import Dict, List, Optional


class CopySpec:
    """Copy specification of a single file or directory."""

    def __init__(self, source_path: str, target_path: Optional[str] = None):
        self.source_path = source_path
        self.target_path = target_path

    def get_target(self) -> str:
        """Get the target path of the copy spec.

        Returns:
            str: If a target_path is given explicitly, it will be returned.
                 Otherwise the source_path will be returned.
        """
        if self.target_path is not None:
            return self.target_path
        return self.source_path


def copy_templates(
    template_dir: str,
    target_dir: str,
    template_file_mapping: List[CopySpec],
    variables: Dict[str, str],
) -> None:
    """Copy templates from the template dir to the target dir.

    Args:
        template_dir (str): Path to the directory containing the template files.
        target_dir (str): Path to the target directory.
        template_file_mapping (Dict[str, str]): A mapping of source path to target path.
        variables (Dict[str, str]): Name to value mapping which will be replaced when parsing the template files.
    """
    for file_to_copy in template_file_mapping:
        with open(
            os.path.join(
                template_dir,
                file_to_copy.source_path,
            ),
            encoding="utf-8",
        ) as file_in:
            target_file_path = os.path.join(target_dir, file_to_copy.get_target())

            os.makedirs(os.path.split(target_file_path)[0], exist_ok=True)
            with open(target_file_path, encoding="utf-8", mode="w") as file_out:
                lines = file_in.readlines()
                for line in lines:
                    for key, value in variables.items():
                        line = line.replace("${{ " + key + " }}", value)
                    file_out.write(line)



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

import os
import subprocess
from io import TextIOWrapper
from typing import List

from velocitas_lib import get_app_manifest, get_workspace_dir, require_env


def build_vehicleapp_image(
    log_output: TextIOWrapper | int = subprocess.DEVNULL,
) -> None:
    """Build VehicleApp docker image and display the progress using a spinner.

    Args:
        log_output (TextIOWrapper | int): Logfile to write or DEVNULL by default.
    """
    app_name = get_app_manifest()["name"].lower()
    image_tag = f"localhost:12345/{app_name}:local"
    dockerfile_path = require_env("dockerfilePath")
    os.environ["DOCKER_BUILDKIT"] = "1"

    extra_proxy_args: List[str] = [
        "--build-arg",
        "HTTP_PROXY",
        "--build-arg",
        "HTTPS_PROXY",
        "--build-arg",
        "FTP_PROXY",
        "--build-arg",
        "ALL_PROXY",
        "--build-arg",
        "NO_PROXY",
    ]

    subprocess.check_call(
        [
            "docker",
            "build",
            "-f",
            dockerfile_path,
            "--progress=plain",
            "-t",
            image_tag,
        ]
        + extra_proxy_args
        + [
            ".",
            "--no-cache",
        ],
        stdout=log_output,
        stderr=log_output,
        cwd=get_workspace_dir(),
    )


def is_docker_image_build_locally(app_name: str) -> bool:
    """Check if vehicle app docker image is locally available

    Args:
        app_name (str): App name to check for
    """
    output = subprocess.check_output(
        [
            "docker",
            "images",
            "-a",
            f"localhost:12345/{app_name}:local",
            "--format",
            "{{.Repository}}:{{.Tag}}",
        ],
    )
    return output.decode("utf-8").strip() == f"localhost:12345/{app_name}:local"


def push_docker_image_to_registry(
    app_name: str, log_output: TextIOWrapper | int = subprocess.DEVNULL
) -> None:
    """Push docker image to local image registry

    Args:
        app_name (str): App name to push to registry
        log_output (TextIOWrapper | int): Logfile to write or DEVNULL by default.
    """
    subprocess.check_call(
        ["docker", "push", f"localhost:12345/{app_name}:local"],
        stdout=log_output,
        stderr=log_output,
    )


def container_exists(
    name: str, log_output: TextIOWrapper | int = subprocess.DEVNULL
) -> bool:
    """Check if a container with a given name exists.

    Args:
        log_output (TextIOWrapper | int): Logfile to write or DEVNULL by default.

    Returns:
        bool: True if the container exists, False if not.
    """
    return "" != str(
        subprocess.check_output(
            ["docker", "ps", "-a", "-q", "-f", f"name={name}"], stderr=log_output
        ),
        "utf-8",
    )



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

import re
from typing import Any, Dict

from velocitas_lib import get_cache_data, get_package_path


def json_obj_to_flat_map(
    obj: Any, prefix: str = "", separator: str = "."
) -> Dict[str, str]:
    """Flatten a JSON Object into a one dimensional dict by joining the keys
    with the specified separator."""
    result = dict[str, str]()
    if isinstance(obj, dict):
        for key, value in obj.items():
            nested_key = f"{prefix}{separator}{key}"
            result.update(json_obj_to_flat_map(value, nested_key, separator))
    elif isinstance(obj, list):
        for index, value in enumerate(obj):
            nested_key = f"{prefix}{separator}{index}"
            result.update(json_obj_to_flat_map(value, nested_key, separator))
    else:
        nested_key = f"{prefix}"
        result[nested_key] = obj

    return result


class ProjectVariables:
    def __init__(self, env: Dict[str, str]):
        self.__build_variables_map(env)

    def __build_variables_map(self, env: Dict[str, str]) -> None:
        variables: Dict[str, str] = {}
        variables.update(json_obj_to_flat_map(get_cache_data(), "builtin.cache"))
        variables.update(env)
        variables["builtin.package_dir"] = get_package_path()
        self._variables = variables

    def replace_occurrences(self, input_str: str) -> str:
        """Replace all occurrences of the defined variables in the input string"""
        if "${{" not in input_str:
            return input_str
        input_str_match = re.search(r"(?<=\${{)(.*?)(?=}})", input_str)
        if input_str_match:
            input_str_value = input_str_match.group().strip()
            if input_str_value not in self._variables:
                raise KeyError(f"{input_str_value!r} not in {self._variables!r}")
            for key, value in self._variables.items():
                input_str = input_str.replace("${{ " + key + " }}", str(value))
            return input_str
        else:
            raise ValueError(f"{input_str!r} not in the right format")



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

# flake8: noqa: U100 unused argument (because of pytest.fixture)

import os

import pytest

from velocitas_lib import get_package_path, get_workspace_dir
from velocitas_lib.services import resolve_functions


@pytest.fixture()
def set_env_vars() -> None:
    os.environ["VELOCITAS_WORKSPACE_DIR"] = "./tests/project"
    os.environ["VELOCITAS_PACKAGE_DIR"] = "./tests/package"


def test_resolve_functions__unknown_function__raises_exception():
    with pytest.raises(RuntimeError, match="Unsupported function: 'foo'!"):
        resolve_functions("$foo( asd )")


def test_resolve_functions__pathInWorkspaceOrPackage__file_missing__raises_exception(
    set_env_vars,
):
    with pytest.raises(
        RuntimeError, match="Path 'asd' not found in workspace or package!"
    ):
        resolve_functions("$pathInWorkspaceOrPackage( asd )")


def test_resolve_functions__pathInWorkspaceOrPackage__file_exists_in_workspace():
    assert resolve_functions("$pathInWorkspaceOrPackage( foo.json )") == os.path.join(
        get_workspace_dir(), "foo.json"
    )


def test_resolve_functions__pathInWorkspaceOrPackage__file_exists_in_package(
    set_env_vars,
):
    assert resolve_functions(
        "$pathInWorkspaceOrPackage( manifest.json )"
    ) == os.path.join(get_package_path(), "manifest.json")



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

import json
import os
import re
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional

from velocitas_lib import get_package_path, get_workspace_dir, require_env
from velocitas_lib.variables import ProjectVariables


class ServiceSpecConfig(NamedTuple):
    image: str
    is_enabled: bool = True
    env_vars: Dict[str, Optional[str]] = dict()
    args: List[str] = list()
    ports: List[str] = list()
    port_forwards: List[str] = list()
    mounts: List[str] = list()
    startup_log_patterns: List[str] = list()


class Service(NamedTuple):
    id: str
    config: ServiceSpecConfig


def resolve_functions(input_str: str) -> str:
    while True:
        input_str_match = re.search(r"\$(\w+)\((.*)\s*\)", input_str)

        if not input_str_match:
            return input_str

        function_name = input_str_match.group(1).strip()
        parameter = input_str_match.group(2).strip()

        return_value = None
        if function_name == "pathInWorkspaceOrPackage":
            path_in_workspace = os.path.join(get_workspace_dir(), parameter)
            if os.path.isfile(path_in_workspace):
                return_value = path_in_workspace

            if return_value is None:
                path_in_package = os.path.join(get_package_path(), parameter)
                if os.path.isfile(path_in_package):
                    return_value = path_in_package

            if return_value is None:
                raise RuntimeError(
                    f"Path {parameter!r} not found in workspace or package!"
                )
        else:
            raise RuntimeError(f"Unsupported function: {function_name!r}!")

        match_span = input_str_match.span(0)
        input_str = (
            input_str[0 : match_span[0]] + return_value + input_str[match_span[1] :]
        )


def parse_service_config(
    service_id: str, service_spec_config: List[Dict[str, Any]]
) -> ServiceSpecConfig:
    """Parse service spec configuration and return it as an named tuple.

    Args:
        service_id: The ID of the service to be parsed.
        service_spec_config: The specificon of the services from config file.
    """

    is_enabled = True
    container_image = None
    env_vars = dict[str, Optional[str]]()
    ports = []
    port_forwards = []
    mounts = []
    args = []
    patterns = []

    variables = ProjectVariables(env=dict(os.environ))

    for config_entry in service_spec_config:
        key = config_entry["key"]
        value = config_entry["value"]

        if isinstance(value, str):
            value = variables.replace_occurrences(value)
            value = resolve_functions(value)

        if key == "enabled":
            is_enabled = value is True or value == "true"
        elif key == "image":
            container_image = value
        elif key == "env":
            pair = value.split("=", 1)
            inner_key = pair[0].strip()
            env_vars[inner_key] = None
            if len(pair) > 1:
                env_vars[inner_key] = pair[1].strip()
        elif key == "arg":
            args.append(value)
        elif key == "port":
            ports.append(value)
        elif key == "port-forward":
            port_forwards.append(value)
        elif key == "mount":
            mounts.append(value)
        elif key == "start-pattern":
            patterns.append(value)

    if container_image is None:
        raise KeyError(f"Service {service_id!r} does not provide an image!")

    return ServiceSpecConfig(
        image=container_image,
        is_enabled=is_enabled,
        env_vars=env_vars,
        args=args,
        ports=ports,
        port_forwards=port_forwards,
        mounts=mounts,
        startup_log_patterns=patterns,
    )


def get_services(verbose: bool = True) -> List[Service]:
    """Return all specified services as Python object."""
    path = Path(f"{get_package_path()}/runtime.json")
    variable_value = require_env("runtimeFilePath")

    if variable_value is not None:
        overwritten_path = Path(variable_value)
        if not overwritten_path.is_absolute():
            overwritten_path = Path(get_workspace_dir()).joinpath(overwritten_path)

        if overwritten_path.exists():
            path = overwritten_path

            if verbose:
                print(f"runtime.json path redirected to {path}")

    json_array: List[Dict[str, Any]] = json.load(
        open(
            path,
            encoding="utf-8",
        )
    )

    services: List[Service] = list()
    for service_json in json_array:
        service_id = service_json["id"]
        service_config = None
        is_service_enabled = True
        if "config" in service_json:
            service_config = parse_service_config(service_id, service_json["config"])
            is_service_enabled = service_config.is_enabled

        if is_service_enabled:
            if service_config is None:
                raise KeyError(f"Service {service_id!r} does not have a config entry!")

            services.append(Service(service_id, service_config))

    return services


def get_specific_service(service_id: str) -> Service:
    """Return the specified service as Python object.

    Args:
        service_id: The ID of the service to be parsed.
    """
    services = get_services()
    services = list(filter(lambda service: service.id == service_id, services))
    if len(services) == 0:
        raise RuntimeError(f"Service with id '{service_id}' not defined")
    if len(services) > 1:
        raise RuntimeError(
            f"Multiple service definitions of id '{service_id}' found, which to take?"
        )
    return services[0]


def get_service_port(service_id: str) -> str:
    """Return the service port as string for the specified service.

    Args:
        service_id: The ID of the service to be parsed.
    """
    return get_specific_service(service_id).config.ports[0]



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0


from typing import Any, Dict, List

from velocitas_lib import get_app_manifest


def get_interfaces_for_type(type: str) -> List[Dict[str, Any]]:
    """Return all interfaces for the given type.

    Args:
        type (str): The type string to look up.

    Returns:
        List: A list of interfaces of the given type.
    """
    manifest_data = get_app_manifest()

    interfaces = []
    for interface in manifest_data["interfaces"]:
        if interface["type"] == type:
            interfaces.append(interface)

    return interfaces



# Copyright (c) 2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

from velocitas_lib.text_utils import (
    replace_item_in_list,
    replace_text_area,
)

test_text_literal = """
class SeatsServiceStub(seats_service_sdk.seats_pb2_grpc.SeatsServicer):
    # <auto-generated>
    '''
    @brief Seats service for getting and controlling the positions of the seats and their
    components in the vehicle.
    This definition corresponds to the COVESA Vehicle Service Catalog (VSC) comfort
    seats service definition (https://github.com/COVESA/vehicle_service_catalog)
    '''

    def Move(self, request, context):
        ''' Set the desired seat position

        Returns gRPC status codes:
        * OK - Seat movement started
        * OUT_OF_RANGE - The addressed seat is not present in this vehicle
        * INVALID_ARGUMENT - At least one of the requested component positions is invalid
        * INTERNAL - A seat service internal error happened - see error message for details
        '''
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

     # </auto-generated>
""".splitlines()

test_text: list[str] = [str(line) for line in test_text_literal]


def test_item_in_list__replaces_all_occurrences_of_item_with_replacement():
    replaced_text = replace_item_in_list(
        test_text,
        "Seats",
        "ReplacedTestString",  # Case sensitive
    )
    matching_count = replaced_text.count("ReplacedTestString")

    assert matching_count == 2


def test_item_in_list__remove_empty_enabled__removes_empty_lines_if_replacement_empty():
    replaced_text = replace_item_in_list(
        test_text,
        "'''",
        remove_empty=True,
    )

    assert len(replaced_text) == 20


def test_item_in_list__remove_empty_disabled__keeps_empty_lines():
    replaced_text = replace_item_in_list(
        test_text,
        "'''",
    )

    assert len(replaced_text) == 24


def test_replace_text_area__replaces_all_text_in_area_with_replacement():
    replaced_text = replace_text_area(
        test_text,
        "# <auto-generated>",
        "# </auto-generated>",
        "ReplacedTestString",
    )
    matching_count = replaced_text.count("ReplacedTestString")

    assert matching_count == 1


def test_replace_text_area__no_replacement_given__removes_lines():
    replaced_text = replace_text_area(
        test_text,
        "'''",
        "'''",
    )

    assert len(replaced_text) == 10



# Copyright (c) 2023-2024 Contributors to the Eclipse Foundation
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0

# flake8: noqa: U100 unused argument (because of pytest.fixture)

import json
import os
import sys
from pathlib import Path

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem

from velocitas_lib import (
    get_app_manifest,
    get_cache_data,
    obtain_local_file_path,
    get_package_path,
    get_script_path,
    get_workspace_dir,
    require_env,
    get_project_cache_dir,
)
from velocitas_lib.services import get_services


@pytest.fixture()
def set_test_env_var() -> str:
    os.environ["TEST"] = "test"
    return os.environ["TEST"]


@pytest.fixture()
def set_velocitas_workspace_dir() -> str:
    os.environ["VELOCITAS_WORKSPACE_DIR"] = "/test/vehicle-app-workspace"
    return os.environ["VELOCITAS_WORKSPACE_DIR"]


@pytest.fixture()
def set_velocitas_package_dir() -> str:
    os.environ["VELOCITAS_PACKAGE_DIR"] = "./tests/package"
    return os.environ["VELOCITAS_PACKAGE_DIR"]


@pytest.fixture()
def set_velocitas_cache_dir() -> str:
    os.environ["VELOCITAS_CACHE_DIR"] = "/test/cache"
    return os.environ["VELOCITAS_CACHE_DIR"]


@pytest.fixture()
def set_app_manifest() -> str:
    app_manifest = {"vehicleModel": {"src": "test"}}
    os.environ["VELOCITAS_APP_MANIFEST"] = json.dumps(app_manifest)
    return os.environ["VELOCITAS_APP_MANIFEST"]


@pytest.fixture()
def set_velocitas_cache_data() -> str:
    cache_data_mock = {"testPropA": "testValueA", "testPropB": "testValueB"}
    os.environ["VELOCITAS_CACHE_DATA"] = json.dumps(cache_data_mock)
    return os.environ["VELOCITAS_CACHE_DATA"]


@pytest.fixture()
def mock_filesystem(fs: FakeFilesystem) -> FakeFilesystem:
    fs.add_real_file(
        os.path.join(Path(__file__).resolve().parent, "package", "manifest.json")
    )
    return fs


def test_require_env__env_var_set__returns_env_value(set_test_env_var):  # type: ignore
    assert require_env("TEST") == "test"


def test_require_env__env_var_not_set__raises_ValueError():
    with pytest.raises(ValueError):
        require_env("TEST_ENV_NOT_SET")


def test_get_workspace_dir__returns_workspace_dir_path(set_velocitas_workspace_dir):  # type: ignore
    assert get_workspace_dir() == "/test/vehicle-app-workspace"


def test_get_app_manifest__app_manifest_set__returns_app_manifest_data(
    set_app_manifest,
):
    assert get_app_manifest()["vehicleModel"]["src"] == "test"


def test_get_app_manifest__missing_key__raises_KeyError(set_app_manifest):  # type: ignore
    with pytest.raises(KeyError):
        get_app_manifest()["vehicleModel"]["srcs"]


def test_get_app_manifest__no_app_manifest__raises_ValueError():
    os.environ["VELOCITAS_APP_MANIFEST"] = ""
    with pytest.raises(ValueError):
        get_app_manifest()


def test_get_script_path__returns_script_path():
    assert get_script_path() == os.path.dirname(os.path.realpath(sys.argv[0]))


def test_get_package_path__returns_package_path(set_velocitas_package_dir):
    assert get_package_path() == "./tests/package"


def test_get_cache_data__returns_cache_data(set_velocitas_cache_data):  # type: ignore
    assert get_cache_data()["testPropA"] == "testValueA"
    assert get_cache_data()["testPropB"] == "testValueB"


def test_get_services__no_overwrite_provided__returns_default_services(
    set_velocitas_package_dir,
    mock_filesystem: FakeFilesystem,
):
    os.environ["runtimeFilePath"] = "runtime.json"
    mock_filesystem.create_file(
        f"{get_package_path()}/runtime.json",
        contents='[ { "id": "service1", "config": [ { "key": "image", "value": "image-service1" } ] } ]',
    )

    all_services = get_services()

    assert len(all_services) == 1
    assert all_services[0].id == "service1"
    assert all_services[0].config.image == "image-service1"


def test_get_services__overwrite_provided__returns_overwritten_services(
    set_velocitas_package_dir,
    mock_filesystem: FakeFilesystem,
):
    os.environ["runtimeFilePath"] = "runtime.json"

    mock_filesystem.create_file(
        f"{get_package_path()}/runtime.json",
        contents='[ { "id": "service1", "config": [ { "key": "image", "value": "image-service1" } ] } ]',
    )
    mock_filesystem.create_file(
        f"{get_workspace_dir()}/runtime.json",
        contents='[ { "id": "my-custom-service", "config": [ { "key": "image", "value": "image-my-custom-service" } ] } ]',
    )

    all_services = get_services()

    assert len(all_services) == 1
    assert all_services[0].id == "my-custom-service"
    assert all_services[0].config.image == "image-my-custom-service"

    mock_filesystem.reset()


def test_obtain_local_file_path__absolute_local_path(set_velocitas_workspace_dir):
    root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    assert obtain_local_file_path(f"{root}/README.md") == f"{root}/README.md"


def test_obtain_local_file_path__relative_local_path(set_velocitas_workspace_dir):
    assert obtain_local_file_path("README.md") == "README.md"


def test_obtain_local_file_path__absolute_local_path_not_available(
    set_velocitas_workspace_dir,
):
    pytest.raises(
        FileNotFoundError,
        obtain_local_file_path,
        "/workspaces/velocitas-lib/README2.md",
    )


def test_obtain_local_file_path__relative_local_path_not_available(
    set_velocitas_workspace_dir,
):
    pytest.raises(FileNotFoundError, obtain_local_file_path, "README2.md")


def test_obtain_local_file_path__uri(set_velocitas_workspace_dir):
    assert (
        obtain_local_file_path(
            "https://raw.githubusercontent.com/eclipse-velocitas/velocitas-lib/main/README.md",
            "/workspaces/velocitas-lib/.pytest_cache/README.md",
        )
        == "/workspaces/velocitas-lib/.pytest_cache/README.md"
    )
    Path.unlink(Path("/workspaces/velocitas-lib/.pytest_cache/README.md"))


def test_obtain_local_file_path__uri_no_download_path(
    set_velocitas_cache_dir,
    mock_filesystem: FakeFilesystem,
):
    mock_filesystem.create_dir(get_project_cache_dir())
    assert (
        obtain_local_file_path(
            "https://raw.githubusercontent.com/eclipse-velocitas/velocitas-lib/main/README.md"
        )
        == "/test/cache/downloads/README.md"
    )
